import pathlib
import shutil
import tempfile


data_path = pathlib.Path(__file__).parent / "data"


def get_clean_data_path():
    """Return .rtdc file in a temp dir for modification"""
    tdir = tempfile.mkdtemp(prefix="dctag_data_")
    orig = data_path / "blood_rbc_leukocytes.rtdc"
    new = pathlib.Path(tdir) / orig.name
    shutil.copy2(orig, new)
    return new


def get_raw_string(some_string):
    return f"{some_string}".encode('unicode_escape').decode()



import pyqtgraph as pg


class SimplePlotItem(pg.PlotItem):
    """Custom class for data visualization in DCTag

    Modifications include:
    - top and right axes
    """

    def __init__(self, parent=None, *args, **kwargs):
        super(SimplePlotItem, self).__init__(parent, *args, **kwargs)
        # show top and right axes, but not ticklabels
        for kax in ["top", "right"]:
            self.showAxis(kax)
            ax = self.axes[kax]["item"]
            ax.setTicks([])
            ax.setLabel(None)
            ax.setStyle(tickTextOffset=0,
                        tickTextWidth=0,
                        tickTextHeight=0,
                        autoExpandTextSpace=False,
                        showValues=False,
                        )
        # This screwed up zooming with right-click:
        # self.showGrid(x=True, y=True, alpha=.1)
        # Use the new GridItem instead:
        grid = pg.GridItem()
        grid.setTextPen("#FFFF")  # hide grid tick labels
        self.addItem(grid)

        # visualization
        self.hideButtons()


class ScatterPlotWidget(pg.PlotWidget):
    """Custom class for data visualization in DCTag
    """

    def __init__(self, parent=None, background='w', **kargs):
        plot_item = SimplePlotItem(**kargs)
        super(ScatterPlotWidget, self).__init__(parent,
                                                background=background,
                                                plotItem=plot_item)

        self.scatter = RTDCScatterPlot()
        self.addItem(self.scatter)
        self.select = pg.PlotDataItem(x=[1], y=[2], symbol="o",
                                      symbolBrush="red")
        self.select.hide()
        self.addItem(self.select)

    def set_scatter(self, x, y):
        self.scatter.setData(x, y)

    def set_event(self, x, y):
        self.select.show()
        self.select.setData([x], [y])


class RTDCScatterPlot(pg.ScatterPlotItem):
    def __init__(self, size=3, pen=None, brush=None, *args, **kwargs):
        if pen is None:
            pen = pg.mkPen(color=(0, 0, 0, .5))
        if brush is None:
            brush = pg.mkBrush("k")
        super(RTDCScatterPlot, self).__init__(size=size,
                                              pen=pen,
                                              brush=brush,
                                              symbol="s",
                                              *args,
                                              **kwargs)
        self.setData(x=range(10), y=range(10), brush=brush)



import pytest

import dclab
import h5py
import numpy as np

from dctag import session

from .helper import get_clean_data_path, get_raw_string


def test_basic():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        assert dts.event_count == 18
        lock_path = dts.path_lock
        assert lock_path.exists()
    assert not lock_path.exists()


def test_is_dctag_session():
    path = get_clean_data_path()
    assert not session.is_dctag_session(path)
    with session.DCTagSession(path, "Peter"):
        pass
    assert session.is_dctag_session(path)


def test_flush_with_missing_file_error():
    path = get_clean_data_path()
    # error should be raised on flush and on __exit__
    with pytest.raises(session.DCTagSessionWriteError,
                       match=get_raw_string(path)):
        with session.DCTagSession(path, "Peter") as dts:
            dts.set_score("ml_score_abc", 0, True)
            path.unlink()
            with pytest.raises(session.DCTagSessionWriteError,
                               match=get_raw_string(path)):
                dts.flush()


def test_get_score_basic():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 2, False)
        dts.set_score("ml_score_abc", 1, True)

        assert dts.get_score("ml_score_abc", 0) is True
        assert dts.get_score("ml_score_abc", 1) is True
        assert dts.get_score("ml_score_abc", 2) is False
        assert np.isnan(dts.get_score("ml_score_abc", 3))
        assert np.isnan(dts.get_score("ml_score_ukn", 3))

    # again in new session
    with session.DCTagSession(path, "Peter") as dts:
        assert dts.get_score("ml_score_abc", 0) is True
        assert dts.get_score("ml_score_abc", 1) is True
        assert dts.get_score("ml_score_abc", 2) is False
        assert np.isnan(dts.get_score("ml_score_abc", 3))
        assert np.isnan(dts.get_score("ml_score_ukn", 3))


def test_get_score_linked():
    path = get_clean_data_path()
    with session.DCTagSession(
            path,
            "Peter",
            linked_features=["ml_score_abc", "ml_score_123"]) as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 2, False)
        dts.set_score("ml_score_abc", 1, True)
        dts.set_score("ml_score_123", 3, True)
        dts.set_score("ml_score_000", 4, True)
        dts.set_score("ml_score_123", 5, False)

        assert dts.get_score("ml_score_abc", 0) is True
        assert dts.get_score("ml_score_abc", 1) is True
        assert dts.get_score("ml_score_abc", 2) is False
        assert dts.get_score("ml_score_abc", 3) is False
        assert np.isnan(dts.get_score("ml_score_abc", 4))
        assert np.isnan(dts.get_score("ml_score_abc", 5))

        assert dts.get_score("ml_score_123", 0) is False
        assert dts.get_score("ml_score_123", 1) is False
        assert np.isnan(dts.get_score("ml_score_123", 2))
        assert dts.get_score("ml_score_123", 3) is True
        assert np.isnan(dts.get_score("ml_score_123", 4))
        assert dts.get_score("ml_score_123", 5) is False

        assert np.isnan(dts.get_score("ml_score_000", 0))
        assert np.isnan(dts.get_score("ml_score_000", 1))
        assert np.isnan(dts.get_score("ml_score_000", 2))
        assert np.isnan(dts.get_score("ml_score_000", 3))
        assert dts.get_score("ml_score_000", 4) is True
        assert np.isnan(dts.get_score("ml_score_000", 5))

    # again in new session
    with session.DCTagSession(path, "Peter") as dts:
        assert dts.get_score("ml_score_abc", 0) is True
        assert dts.get_score("ml_score_abc", 1) is True
        assert dts.get_score("ml_score_abc", 2) is False
        assert dts.get_score("ml_score_abc", 3) is False
        assert np.isnan(dts.get_score("ml_score_abc", 4))
        assert np.isnan(dts.get_score("ml_score_abc", 5))

        assert dts.get_score("ml_score_123", 0) is False
        assert dts.get_score("ml_score_123", 1) is False
        assert np.isnan(dts.get_score("ml_score_123", 2))
        assert dts.get_score("ml_score_123", 3) is True
        assert np.isnan(dts.get_score("ml_score_123", 4))
        assert dts.get_score("ml_score_123", 5) is False

        assert np.isnan(dts.get_score("ml_score_000", 0))
        assert np.isnan(dts.get_score("ml_score_000", 1))
        assert np.isnan(dts.get_score("ml_score_000", 2))
        assert np.isnan(dts.get_score("ml_score_000", 3))
        assert dts.get_score("ml_score_000", 4) is True
        assert np.isnan(dts.get_score("ml_score_000", 5))


def test_log_basic():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 1, True)
        dts.set_score("ml_score_abc", 0, False)

    expected = [
        "user: Peter",
        "",
        "New session with DCTag ",
        "Linked features: []",
        "ml_score_abc count False: 1",
        "ml_score_abc count True: 2",
    ]

    with dclab.new_dataset(path) as ds:
        for line, exp in zip(ds.logs["dctag-history"], expected):
            assert exp in line


def test_log_linked_features():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 1, True)
        dts.set_score("ml_score_abc", 0, False)
        dts.linked_features = ["ml_score_abc", "ml_score_456"]
        dts.set_score("ml_score_abc", 2, True)
        dts.set_score("ml_score_456", 3, True)

    expected = [
        "user: Peter",
        "",
        "New session with DCTag ",
        "Linked features: []",
        "ml_score_abc count False: 1",
        "ml_score_abc count True: 2",
        "",
        "New session with DCTag ",
        "Linked features: ['ml_score_456', 'ml_score_abc']",
        "ml_score_456 count True: 1",
        "ml_score_abc count True: 1",
    ]

    with dclab.new_dataset(path) as ds:
        for line, exp in zip(ds.logs["dctag-history"], expected):
            assert exp in line


def test_log_user():
    """At the beginning, onle the user should be written"""
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter"):
        pass
    with dclab.new_dataset(path) as ds:
        assert "".join(ds.logs["dctag-history"]).strip() == "user: Peter"


def test_reset_score_lists_and_history():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 1, False)
        dts.set_score("ml_score_abc", 2, True)

        dts.reset_score("ml_score_abc", 0)
        dts.reset_score("ml_score_abc", 1)

        assert dts.scores[0] == ("ml_score_abc", 0, True)
        assert dts.scores[1] == ("ml_score_abc", 1, False)
        assert dts.scores[2] == ("ml_score_abc", 2, True)
        assert dts.scores[3] == ("ml_score_abc", 0, np.nan)
        assert dts.scores[4] == ("ml_score_abc", 1, np.nan)

    # now check the data file to see that this worked
    with dclab.new_dataset(path) as ds:
        assert np.isnan(ds["ml_score_abc"][0])
        assert np.isnan(ds["ml_score_abc"][1])
        assert ds["ml_score_abc"][2] == 1
        assert np.all(np.isnan(ds["ml_score_abc"][3:]))

        # now check that the logs were written
        assert "dctag-history" in ds.logs
        dctaglog = "\n".join(ds.logs["dctag-history"])
        assert "ml_score_abc count True: 2" in dctaglog
        assert "ml_score_abc count False: 1" in dctaglog
        assert "ml_score_abc count reset: 2" in dctaglog
        assert dctaglog.startswith("user: Peter")


def test_reset_score_with_linked_features():
    path = get_clean_data_path()
    linked = ["ml_score_001", "ml_score_002"]
    with session.DCTagSession(path, "Peter", linked_features=linked) as dts:
        dts.set_score("ml_score_001", 0, True)
        dts.set_score("ml_score_ot1", 0, False)

        dts.set_score("ml_score_ot1", 1, True)
        dts.set_score("ml_score_ot2", 1, False)
        dts.set_score("ml_score_002", 1, True)

        dts.reset_score("ml_score_001", 0)
        dts.reset_score("ml_score_ot1", 1)

    with dclab.new_dataset(path) as ds:
        assert np.isnan(ds["ml_score_001"][0])
        assert np.isnan(ds["ml_score_002"][0])
        assert ds["ml_score_ot1"][0] == 0
        assert np.isnan(ds["ml_score_ot2"][0])

        assert ds["ml_score_001"][1] == 0
        assert ds["ml_score_002"][1] == 1
        assert np.isnan(ds["ml_score_ot1"][1])
        assert ds["ml_score_ot2"][1] == 0


def test_set_score_lists_and_history():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 2, False)
        dts.set_score("ml_score_abc", 1, True)
        dts.set_score("ml_score_abd", 3, True)
        assert dts.scores[0] == ("ml_score_abc", 0, True)
        assert dts.scores[1] == ("ml_score_abc", 2, False)
        assert dts.scores[2] == ("ml_score_abc", 1, True)
        assert dts.scores[3] == ("ml_score_abd", 3, True)
        assert dts.history["ml_score_abc count True"] == 2
        assert dts.history["ml_score_abc count False"] == 1
        assert dts.history["ml_score_abd count True"] == 1

    # now check the data file to see that this worked
    with dclab.new_dataset(path) as ds:
        assert ds["ml_score_abc"][0] == 1
        assert ds["ml_score_abc"][1] == 1
        assert ds["ml_score_abc"][2] == 0
        assert np.all(np.isnan(ds["ml_score_abc"][3:]))
        assert np.isnan(ds["ml_score_abd"][0])
        assert np.isnan(ds["ml_score_abd"][1])
        assert np.isnan(ds["ml_score_abd"][2])
        assert ds["ml_score_abd"][3] == 1
        assert np.all(np.isnan(ds["ml_score_abd"][4:]))

        # now check that the logs were written
        assert "dctag-history" in ds.logs
        dctaglog = "\n".join(ds.logs["dctag-history"])
        assert "ml_score_abc count True: 2" in dctaglog
        assert dctaglog.startswith("user: Peter")


def test_set_score_multiple_ratings_for_index():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 2, False)
        dts.set_score("ml_score_abc", 0, False)

    with dclab.new_dataset(path) as ds:
        assert ds["ml_score_abc"][0] == 0
        assert np.isnan(ds["ml_score_abc"][1])
        assert ds["ml_score_abc"][2] == 0
        assert np.all(np.isnan(ds["ml_score_abc"][3:]))

        # now check that the logs were written
        assert "dctag-history" in ds.logs
        dctaglog = "\n".join(ds.logs["dctag-history"])
        assert "ml_score_abc count True: 1" in dctaglog
        assert "ml_score_abc count False: 2" in dctaglog


def test_set_score_with_linked_features():
    path = get_clean_data_path()
    linked = ["ml_score_001", "ml_score_002"]
    with session.DCTagSession(path, "Peter", linked_features=linked) as dts:
        dts.set_score("ml_score_001", 0, True)
        dts.set_score("ml_score_ot1", 0, False)

        dts.set_score("ml_score_ot1", 1, True)
        dts.set_score("ml_score_ot2", 1, False)
        dts.set_score("ml_score_002", 1, True)

        dts.set_score("ml_score_001", 2, True)
        dts.set_score("ml_score_002", 2, False)

        dts.set_score("ml_score_002", 3, True)
        dts.set_score("ml_score_001", 3, True)

        dts.set_score("ml_score_ot1", 4, True)
        dts.set_score("ml_score_001", 4, False)

    with dclab.new_dataset(path) as ds:
        assert ds["ml_score_001"][0] == 1
        assert ds["ml_score_002"][0] == 0
        assert ds["ml_score_ot1"][0] == 0
        assert np.isnan(ds["ml_score_ot2"][0])

        assert ds["ml_score_001"][1] == 0
        assert ds["ml_score_002"][1] == 1
        assert ds["ml_score_ot1"][1] == 1
        assert ds["ml_score_ot2"][1] == 0

        assert ds["ml_score_001"][2] == 1
        assert ds["ml_score_002"][2] == 0
        assert np.isnan(ds["ml_score_ot1"][2])
        assert np.isnan(ds["ml_score_ot2"][2])

        assert ds["ml_score_001"][3] == 1
        assert ds["ml_score_002"][3] == 0
        assert np.isnan(ds["ml_score_ot1"][3])
        assert np.isnan(ds["ml_score_ot2"][3])

        assert ds["ml_score_001"][4] == 0
        assert np.isnan(ds["ml_score_002"][4])
        assert ds["ml_score_ot1"][4] == 1
        assert np.isnan(ds["ml_score_ot2"][4])


def test_set_score_wrong_feature_error():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        with pytest.raises(ValueError, match="Expected 'ml_score_xxx' or"):
            dts.set_score("volume", 0, True)
        with pytest.raises(ValueError, match="Expected 'ml_score_xxx' or"):
            dts.set_score("ml_flore_abc", 0, True)


def test_session_autocomplete_linked_features():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_001", 0, True)
        dts.set_score("ml_score_ot1", 0, False)
        dts.set_score("ml_score_ot1", 1, True)
        dts.set_score("ml_score_ot2", 1, False)
        dts.set_score("ml_score_002", 1, True)

        # sanity checks
        assert dts.get_score("ml_score_001", 0) is True
        assert dts.get_score("ml_score_ot1", 0) is False
        assert dts.get_score("ml_score_ot1", 1) is True
        assert dts.get_score("ml_score_ot2", 1) is False
        assert dts.get_score("ml_score_002", 1) is True

        # to be tested after linking and autocompletion
        assert np.isnan(dts.get_score("ml_score_001", 1))
        assert np.isnan(dts.get_score("ml_score_002", 0))
        # perform linking and autocompletion
        dts.linked_features = ["ml_score_001", "ml_score_002"]
        dts.autocomplete_linked_features()

        # sanity checks
        assert dts.get_score("ml_score_001", 0) is True
        assert dts.get_score("ml_score_ot1", 0) is False
        assert dts.get_score("ml_score_ot1", 1) is True
        assert dts.get_score("ml_score_ot2", 1) is False
        assert dts.get_score("ml_score_002", 1) is True
        # more sanity checks
        assert np.isnan(dts.get_score("ml_score_ot2", 0))
        assert np.isnan(dts.get_score("ml_score_ot2", 3))
        # new tests
        assert dts.get_score("ml_score_001", 1) is False
        assert dts.get_score("ml_score_002", 0) is False

    with dclab.new_dataset(path) as ds:
        assert ds["ml_score_001"][0] == 1
        assert ds["ml_score_002"][0] == 0
        assert ds["ml_score_ot1"][0] == 0
        assert np.isnan(ds["ml_score_ot2"][0])

        assert ds["ml_score_001"][1] == 0
        assert ds["ml_score_002"][1] == 1
        assert ds["ml_score_ot1"][1] == 1
        assert ds["ml_score_ot2"][1] == 0


def test_session_autocomplete_linked_features_error():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_001", 0, True)
        dts.set_score("ml_score_002", 0, True)
        dts.linked_features = ["ml_score_001", "ml_score_002"]
        with pytest.raises(
                ValueError,
                match="always only one of those scores is labeled as True"):
            dts.autocomplete_linked_features()


def test_session_backup_scores():
    path = get_clean_data_path()
    linked = ["ml_score_001", "ml_score_002"]
    # We cannot use the context manager, because closing it will raise
    # an exception.
    dts = session.DCTagSession(path, "Peter", linked_features=linked)
    dts.set_score("ml_score_001", 0, True)
    dts.set_score("ml_score_ot1", 0, False)

    dts.set_score("ml_score_ot1", 1, True)
    dts.set_score("ml_score_ot2", 1, False)
    dts.set_score("ml_score_002", 1, True)

    dts.set_score("ml_score_001", 2, True)
    dts.set_score("ml_score_002", 2, False)

    dts.set_score("ml_score_002", 3, True)
    dts.set_score("ml_score_001", 3, True)

    dts.set_score("ml_score_ot1", 4, True)
    dts.set_score("ml_score_001", 4, False)

    # simulate worst-case scenario
    path.unlink()
    assert not path.exists()

    # now create a backup
    backup_path = path.with_name("scores.h5")
    dts.backup_scores(backup_path)

    # now recreate the session by using the same path
    path2 = get_clean_data_path()
    with h5py.File(backup_path) as h5, dclab.RTDCWriter(path2) as hw:
        for feat in h5:
            hw.store_feature(feat, h5[feat])

    # This is the same test as above, only this time from the recreated file
    with dclab.new_dataset(path2) as ds:
        assert ds["ml_score_001"][0] == 1
        assert ds["ml_score_002"][0] == 0
        assert ds["ml_score_ot1"][0] == 0
        assert np.isnan(ds["ml_score_ot2"][0])

        assert ds["ml_score_001"][1] == 0
        assert ds["ml_score_002"][1] == 1
        assert ds["ml_score_ot1"][1] == 1
        assert ds["ml_score_ot2"][1] == 0

        assert ds["ml_score_001"][2] == 1
        assert ds["ml_score_002"][2] == 0
        assert np.isnan(ds["ml_score_ot1"][2])
        assert np.isnan(ds["ml_score_ot2"][2])

        assert ds["ml_score_001"][3] == 1
        assert ds["ml_score_002"][3] == 0
        assert np.isnan(ds["ml_score_ot1"][3])
        assert np.isnan(ds["ml_score_ot2"][3])

        assert ds["ml_score_001"][4] == 0
        assert np.isnan(ds["ml_score_002"][4])
        assert ds["ml_score_ot1"][4] == 1
        assert np.isnan(ds["ml_score_ot2"][4])


def test_session_bool():
    path = get_clean_data_path()
    dts = session.DCTagSession(path, "Peter")
    assert dts
    dts.close()
    assert not dts


def test_session_claim_path_missing_history():
    """Handle missing history properly"""
    path = get_clean_data_path()
    with dclab.RTDCWriter(path) as hw:
        hw.store_log("dctag-history", "Nothing")

    with session.DCTagSession(path, "Peter"):
        pass

    with dclab.new_dataset(path) as ds:
        assert ds.logs["dctag-history"][0] == "user: Peter"


def test_session_claim_path_missing_history_empty_log():
    """Handle missing history properly"""
    path = get_clean_data_path()
    with dclab.RTDCWriter(path) as hw:
        hw.store_log("dctag-history", [])

    with session.DCTagSession(path, "Peter"):
        pass

    with dclab.new_dataset(path) as ds:
        assert ds.logs["dctag-history"][0] == "user: Peter"


def test_session_error_closed_flush():
    path = get_clean_data_path()
    dts = session.DCTagSession(path, "Peter")
    dts.set_score("ml_score_001", 10, False)
    # Force this scenario which otherwise could only be triggered maybe
    # via threading.
    dts._closed = True
    with pytest.raises(session.DCTagSessionClosedError,
                       match="flush the session"):
        dts.flush()


def test_session_error_closed_set_score():
    path = get_clean_data_path()
    dts = session.DCTagSession(path, "Peter")
    dts.close()
    with pytest.raises(session.DCTagSessionClosedError,
                       match="set the score"):
        dts.set_score("ml_score_001", 0, True)


def test_session_error_locked():
    path = get_clean_data_path()
    lock_path = path.with_suffix(".dctag")
    lock_path.touch()
    # make sure the session cannot be opened if it is locked
    with pytest.raises(session.DCTagSessionLockedError,
                       match=get_raw_string(path)):
        with session.DCTagSession(path, "Peter"):
            pass
    # make sure the lock file is not removed by context manager
    assert lock_path.exists()


def test_session_error_wronguser():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter"):
        pass

    with pytest.raises(session.DCTagSessionWrongUserError, match="Peter"):
        with session.DCTagSession(path, "Hans"):
            pass


def test_session_error_wronguser_override():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter"):
        pass

    with session.DCTagSession(path, "Hans", override_user=True):
        pass

    with h5py.File(path) as h5:
        assert h5["logs/dctag-history"][0].decode("utf8") == "user: Hans"
        for line in h5["logs/dctag-history"][:]:
            line = line.decode("utf-8")
            if line.count("Session force-claimed from Peter by Hans."):
                break
        else:
            assert False, "session-claim string missing in log"


def test_session_get_scores_true_basic():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 2, True)
        dts.set_score("ml_score_abc", 3, True)
        dts.set_score("ml_score_456", 3, True)

        assert dts.get_scores_true(0) == ["ml_score_abc"]
        assert dts.get_scores_true(1) == []
        assert dts.get_scores_true(2) == ["ml_score_abc"]
        assert dts.get_scores_true(3) == ["ml_score_456", "ml_score_abc"]
        assert dts.get_scores_true(4) == []


def test_session_get_scores_true_basic_userdef():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("userdef1", 0, True)
        dts.set_score("userdef1", 2, True)
        dts.set_score("userdef1", 3, True)
        dts.set_score("userdef2", 3, True)

        assert dts.get_scores_true(0) == ["userdef1"]
        assert dts.get_scores_true(1) == []
        assert dts.get_scores_true(2) == ["userdef1"]
        assert dts.get_scores_true(3) == ["userdef1", "userdef2"]
        assert dts.get_scores_true(4) == []


def test_session_get_scores_true_linked():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 1, True)
        dts.set_score("ml_score_abc", 0, False)
        dts.linked_features = ["ml_score_abc", "ml_score_456"]
        dts.set_score("ml_score_abc", 2, True)
        dts.set_score("ml_score_456", 3, True)

        assert dts.get_scores_true(0) == []
        assert dts.get_scores_true(1) == ["ml_score_abc"]
        assert dts.get_scores_true(2) == ["ml_score_abc"]
        assert dts.get_scores_true(3) == ["ml_score_456"]


def test_session_multiple_with_linked_features():
    path = get_clean_data_path()
    with session.DCTagSession(path, "Peter") as dts:
        dts.set_score("ml_score_abc", 0, True)
        dts.set_score("ml_score_abc", 1, True)
        dts.set_score("ml_score_abc", 0, False)
        dts.linked_features = ["ml_score_abc", "ml_score_456"]
        dts.set_score("ml_score_abc", 2, True)
        dts.set_score("ml_score_456", 3, True)

    with dclab.new_dataset(path) as ds:
        assert ds["ml_score_abc"][0] == 0
        assert ds["ml_score_abc"][1] == 1
        assert ds["ml_score_abc"][2] == 1
        assert ds["ml_score_abc"][3] == 0

        # unless a method is implemented that does "autocompletion",
        # we should not touch these.
        assert np.isnan(ds["ml_score_456"][0])
        assert np.isnan(ds["ml_score_456"][1])
        # from linked session:
        assert ds["ml_score_456"][2] == 0
        assert ds["ml_score_456"][3] == 1


def test_session_warning_closed_get_score():
    path = get_clean_data_path()
    dts = session.DCTagSession(path, "Peter")
    dts.set_score("ml_score_123", 0, True)
    dts.close()
    with pytest.warns(session.DCTagSessionClosedWarning,
                      match="get the score"):
        assert dts.get_score("ml_score_123", 0)


def test_session_warning_closed_close():
    path = get_clean_data_path()
    dts = session.DCTagSession(path, "Peter")
    dts.set_score("ml_score_123", 0, True)
    dts.close()
    with pytest.warns(session.DCTagSessionClosedWarning,
                      match="close the session"):
        with pytest.warns(session.DCTagSessionClosedWarning,
                          match="flush the session"):
            dts.close()


def test_session_warning_closed_flush():
    path = get_clean_data_path()
    dts = session.DCTagSession(path, "Peter")
    dts.set_score("ml_score_123", 0, True)
    dts.close()
    with pytest.warns(session.DCTagSessionClosedWarning,
                      match="flush the session"):
        dts.flush()



import functools
import importlib.resources

import dclab
import numpy as np
from PyQt5 import QtCore, QtWidgets, uic
import pyqtgraph as pg
from scipy.ndimage import binary_erosion


#: dictionary with default axes limits for these features
LIMITS_FEAT = {
    "deform": [0.0, 0.5],
    "area_um": [10, 130],
    "bright_avg": [70, 140],
    "bright_sd": [5, 35],
    "time": None,
}

#: list with scatter plot axis features (there are three)
SCATTER_FEAT = [
    ["area_um", "deform"],
    ["area_um", "bright_avg"],
    ["bright_sd", "bright_avg"],
    ["time", "bright_avg"],
]


class WidgetVisualize(QtWidgets.QWidget):
    """Widget for visualizing data"""

    def __init__(self, *args, **kwargs):
        super(WidgetVisualize, self).__init__(*args, **kwargs)

        ref = importlib.resources.files("dctag.gui") / "widget_vis.ui"
        with importlib.resources.as_file(ref) as path_ui:
            uic.loadUi(path_ui, self)

        self.session = None

        self.scatter_plots = [self.scatter_1, self.scatter_2, self.scatter_3,
                              self.scatter_4]
        # Set individual plots
        kw0 = dict(x=np.arange(10), y=np.arange(10))
        self.trace_plots = {
            "fl1_raw": pg.PlotDataItem(pen="#15BF00", **kw0),  # green
            "fl2_raw": pg.PlotDataItem(pen="#BF8A00", **kw0),  # orange
            "fl3_raw": pg.PlotDataItem(pen="#BF0C00", **kw0),  # red
            "fl1_median": pg.PlotDataItem(pen="#15BF00", **kw0),  # green
            "fl2_median": pg.PlotDataItem(pen="#BF8A00", **kw0),  # orange
            "fl3_median": pg.PlotDataItem(pen="#BF0C00", **kw0),  # red
        }
        self.widget_trace.plotItem.setLabels(bottom="Event time [µs]")
        self.legend_trace = self.widget_trace.addLegend(
            offset=(+.001, -.001), labelTextSize='7pt', colCount=1)

        self.trace_plot_legends = {}

        for key in self.trace_plots:
            self.widget_trace.addItem(self.trace_plots[key])
            self.trace_plots[key].hide()

        # linked axes
        for ii, plota in enumerate(self.scatter_plots):
            vba = plota.getViewBox()
            for jj, plotb in enumerate(self.scatter_plots):
                vbb = plotb.getViewBox()
                if ii < jj:
                    if SCATTER_FEAT[ii][0] == SCATTER_FEAT[jj][0]:
                        vba.linkView(vba.XAxis, vbb)
                    if SCATTER_FEAT[ii][1] == SCATTER_FEAT[jj][1]:
                        vba.linkView(vba.YAxis, vbb)

        # signals
        self.checkBox_auto_contrast.stateChanged.connect(
            self.update_image_cropped)
        self.spinBox_contrast_max.valueChanged.connect(
            self.update_image_cropped)
        self.spinBox_contrast_min.valueChanged.connect(
            self.update_image_cropped)

    def reset(self, reset_plots=False):
        """Clear current visualization"""
        # clear the event image cache
        self.get_event_data.cache_clear()
        self.get_feature_data.cache_clear()
        # UI
        self.setEnabled(False)
        self.groupBox_event.setTitle("Event")
        if reset_plots:
            self.image_channel.clear()
            self.image_channel_contour.clear()
            self.image_cropped.clear()
            self.legend_trace.clear()
            for key in self.trace_plots:
                line = self.trace_plots[key]
                line.setData(np.arange(10), np.arange(10))
                self.widget_trace.update()
            for plot in self.scatter_plots:
                plot.set_scatter(np.arange(10), np.arange(10))

    @functools.lru_cache(maxsize=900)
    def get_feature_data(self, feature):
        with dclab.new_dataset(self.session.path) as ds:
            return ds[feature][:]

    @functools.lru_cache(maxsize=50)
    def get_event_data(self, index):
        # is this too slow?
        with dclab.new_dataset(self.session.path) as ds:
            pxs = ds.config["imaging"]["pixel size"]
            data = {"image": ds["image"][index],
                    "mask": ds["mask"][index],
                    "pos_x_px": self.get_feature_data("pos_x")[index] / pxs,
                    }
            for feat in LIMITS_FEAT:
                data[feat] = self.get_feature_data(feat)[index]
        return data

    def set_event(self, session, event_index):
        if self.session is not session:
            self.reset()
            self.session = session
            self.update_scatter_plots()
        if self.session:
            # Programmatically, this is always the case, but for clarity,
            # we use the `if self.session` case.
            self.setEnabled(True)
            self.groupBox_event.setTitle(
                f"Event {event_index + 1} (total {session.event_count}) ")
            self.session = session
            data = self.get_event_data(event_index)
            # Plot the channel images
            # raw image
            self.image_channel.setImage(data["image"])
            # image with contour
            image_contour = get_contour_image(data)
            self.image_channel_contour.setImage(image_contour)
            # cropped image
            image_cropped = get_cropped_image(data)
            self.update_image_cropped(image_cropped)

            # Plot event in the scatter plots
            for plot, [featx, featy] in zip(self.scatter_plots, SCATTER_FEAT):
                plot.set_event(data[featx], data[featy])

            # Add the Fluorescence traces of the event
            self.set_fluorescence_traces(event_index)

    @QtCore.pyqtSlot()
    def update_image_cropped(self, image_cropped=None):
        """Udpate the cropped image on the right

        This handles auto-contrast.
        """
        if image_cropped is None:
            # use the current data
            image_cropped = self.image_cropped.image
        if self.checkBox_auto_contrast.isChecked():
            levels = (image_cropped.min(), image_cropped.max())
        else:
            levels = (self.spinBox_contrast_min.value(),
                      self.spinBox_contrast_max.value())
        self.image_cropped.setImage(image_cropped, autoLevels=False,
                                    levels=levels)
        # make sure levels are shown in UI
        self.spinBox_contrast_min.blockSignals(True)
        self.spinBox_contrast_min.setValue(levels[0])
        self.spinBox_contrast_min.blockSignals(False)
        self.spinBox_contrast_max.blockSignals(True)
        self.spinBox_contrast_max.setValue(levels[1])
        self.spinBox_contrast_max.blockSignals(False)

    def update_scatter_plots(self):
        for plot, [featx, featy] in zip(self.scatter_plots, SCATTER_FEAT):
            with dclab.new_dataset(self.session.path) as ds:
                x, y = ds.get_downsampled_scatter(xax=featx,
                                                  yax=featy,
                                                  downsample=10000)
            plot.set_scatter(x, y)
            if LIMITS_FEAT[featx] is not None:
                plot.setXRange(*LIMITS_FEAT[featx])
            if LIMITS_FEAT[featy] is not None:
                plot.setYRange(*LIMITS_FEAT[featy])
            plot.setLabel('bottom', dclab.dfn.get_feature_label(featx))
            plot.setLabel('left', dclab.dfn.get_feature_label(featy))

    def set_fluorescence_traces(self, event_index):
        """Set the fluorescence traces on the widget"""
        self.legend_trace.clear()
        with dclab.new_dataset(self.session.path) as ds:
            if "trace" in ds:
                self.widget_trace.show()
                # time axis
                fl_samples = ds.config["fluorescence"]["samples per event"]
                fl_rate = ds.config["fluorescence"]["sample rate"]
                fl_time = np.arange(fl_samples) / fl_rate * 1e6
                # temporal range (min, max, fl-peak-maximum)
                range_t = [fl_time[0], fl_time[-1], 0]
                # fluorescence intensity
                range_fl = [0, 0]

                # Use this list to only show one trace type (raw or median)
                shown_traces = []

                for key in dclab.dfn.FLUOR_TRACES:
                    trid = key.split("_")[0]
                    if key in ds["trace"] and trid not in shown_traces:
                        shown_traces.append(trid)
                        # show the trace information
                        tracey = ds["trace"][key][event_index]  # trace data
                        range_fl[0] = min(range_fl[0], tracey.min())
                        range_fl[1] = max(range_fl[1], tracey.max())
                        self.trace_plots[key].setData(fl_time, tracey)
                        self.trace_plots[key].show()
                        # set legend name
                        ln = "{} {}".format(
                            "FL-{}".format(key[2]),
                            'median' if str(key[4]) == 'm' else 'raw')
                        self.legend_trace.addItem(self.trace_plots[key], ln)
                        self.legend_trace.update()
                    else:
                        self.trace_plots[key].hide()
                self.widget_trace.setXRange(*range_t[:2], padding=0)
                if range_fl[0] != range_fl[1]:
                    self.widget_trace.setYRange(*range_fl, padding=.01)
                self.widget_trace.setLimits(xMin=0, xMax=fl_time[-1])
            else:
                self.widget_trace.hide()


def get_contour_image(event_data):
    image = event_data["image"]
    mask = event_data["mask"]
    cellimg = np.copy(image)
    cellimg = cellimg.reshape(
        cellimg.shape[0], cellimg.shape[1], 1)
    cellimg = np.repeat(cellimg, 3, axis=2)
    # clip and convert to int
    cellimg = np.clip(cellimg, 0, 255)
    cellimg = np.require(cellimg, np.uint8, 'C')
    # Compute contour image from mask. If you are wondering
    # whether this is kosher, please take a look at issue #76:
    # https://github.com/ZELLMECHANIK-DRESDEN/dclab/issues/76
    cont = mask ^ binary_erosion(mask)
    # set red contour pixel values in original image
    cellimg[cont, 0] = int(255 * .7)
    cellimg[cont, 1] = 0
    cellimg[cont, 2] = 0
    return cellimg


def get_cropped_image(event_data):
    image = event_data["image"]
    pos_lat = int(event_data["pos_x_px"])
    width = image.shape[0]
    left = max(0, pos_lat - width // 2)
    left = min(left, image.shape[1] - width)
    right = left + width
    cropped = image[:, left:right]
    return cropped



from collections import OrderedDict
import functools
import importlib.resources
import json
import pathlib

import dclab


def get_available_label_groups():
    files = []
    for entry in importlib.resources.files("dctag.resources").iterdir():
        if entry.name.endswith(".json"):
            files.append(entry.name)

    names = [pp[4:-5] for pp in files]
    return sorted(names)


@functools.lru_cache(maxsize=100)
def get_dctag_label_dict(name="ml_scores_blood"):
    fname = f"tag_{name}.json"
    ref = importlib.resources.files("dctag.resources") / fname
    with importlib.resources.as_file(ref) as path:
        score_dict = json.loads(pathlib.Path(path).read_text(),
                                # load as ordered dictionary
                                object_pairs_hook=OrderedDict)
    return score_dict


def get_feature_label(feature, label_group=None):
    default_label = dclab.dfn.get_feature_label(feature)
    if label_group is None:
        # go through all label groups
        for group in get_available_label_groups():
            label = get_feature_label(feature, group)
            if label != default_label:
                break
        else:
            label = default_label
    else:
        # use this specific label group
        score_dict = get_dctag_label_dict(name=label_group)
        if feature in score_dict:
            label = score_dict[feature]["label"]
        else:
            label = default_label
    return label


def get_feature_shortcut(feature, label_group="ml_scores_blood"):
    score_dict = get_dctag_label_dict(name=label_group)
    if feature in score_dict:
        return score_dict[feature]["shortcut"]
    else:
        return "A"



import pathlib

from PyQt5 import QtCore, QtWidgets
import pytest

from dctag import session
from dctag.gui.main import DCTag
from .helper import get_clean_data_path


data_dir = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)
    pass
    # A test function will be run at this point
    yield
    # Code that will run after your test
    # restore dctag-tester for other tests
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setValue("user/name", "dctag-tester")
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)


def test_empty_session(mw):
    QtWidgets.QApplication.setActiveWindow(mw)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    # make sure things are disabled
    assert not mw.tab_binary.isEnabled()


@pytest.mark.parametrize("event_index,expected", [
    [-10, 0],
    [-1, 0],
    [0, 0],
    [16, 16],
    [17, 17],
    [18, 17],
    [5000, 17]])
def test_goto_event_limits(event_index, expected, qtbot, mw):
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    # start labeling
    qtbot.mouseClick(mw.tab_binary.pushButton_start, QtCore.Qt.LeftButton)

    # go to event
    mw.tab_binary.goto_event(event_index)
    assert mw.tab_binary.event_index == expected


def test_goto_event_button_labels(qtbot, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 1, False)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, False)

    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    idx = mw.tab_binary.comboBox_score.findData("ml_score_r1f")
    mw.tab_binary.comboBox_score.setCurrentIndex(idx)

    qtbot.mouseClick(mw.tab_binary.pushButton_start, QtCore.Qt.LeftButton)

    # The first event should be displayed, and it should be set to True
    assert mw.tab_binary.pushButton_yes.text() == "[Yes]"
    assert mw.tab_binary.pushButton_no.text() == "No"
    assert mw.tab_binary.label_score_prev.text() == ""
    assert mw.tab_binary.label_score_next.text() == "No"

    # click on next.
    qtbot.mouseClick(mw.tab_binary.pushButton_next, QtCore.Qt.LeftButton)

    # This should be False now
    assert mw.tab_binary.pushButton_yes.text() == "Yes"
    assert mw.tab_binary.pushButton_no.text() == "[No]"
    assert mw.tab_binary.label_score_prev.text() == "Yes"
    assert mw.tab_binary.label_score_next.text() == "Yes"


def test_goto_event_button_labels_userdef(qtbot):
    settings = QtCore.QSettings()
    settings.setValue("labeling group", "userdef")

    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("userdef1", 0, True)
        dts.set_score("userdef1", 1, False)
        dts.set_score("userdef1", 2, True)
        dts.set_score("userdef1", 3, False)

    mw = DCTag()
    qtbot.addWidget(mw)
    QtWidgets.QApplication.setActiveWindow(mw)
    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    idx = mw.tab_binary.comboBox_score.findData("userdef1")
    mw.tab_binary.comboBox_score.setCurrentIndex(idx)

    qtbot.mouseClick(mw.tab_binary.pushButton_start, QtCore.Qt.LeftButton)

    # The first event should be displayed, and it should be set to True
    assert mw.tab_binary.pushButton_yes.text() == "[Yes]"
    assert mw.tab_binary.pushButton_no.text() == "No"
    assert mw.tab_binary.label_score_prev.text() == ""
    assert mw.tab_binary.label_score_next.text() == "No"

    # click on next.
    qtbot.mouseClick(mw.tab_binary.pushButton_next, QtCore.Qt.LeftButton)

    # This should be False now
    assert mw.tab_binary.pushButton_yes.text() == "Yes"
    assert mw.tab_binary.pushButton_no.text() == "[No]"
    assert mw.tab_binary.label_score_prev.text() == "Yes"
    assert mw.tab_binary.label_score_next.text() == "Yes"
    mw.close()
    settings.setValue("labeling group", "ml_scores_blood")


def test_event_push_buttons(qtbot, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 1, False)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, False)

    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    idx = mw.tab_binary.comboBox_score.findData("ml_score_r1f")
    mw.tab_binary.comboBox_score.setCurrentIndex(idx)

    qtbot.mouseClick(mw.tab_binary.pushButton_start, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 0
    qtbot.mouseClick(mw.tab_binary.pushButton_next, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 1
    qtbot.mouseClick(mw.tab_binary.pushButton_fast_next, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 4
    qtbot.mouseClick(mw.tab_binary.pushButton_prev, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 3
    qtbot.mouseClick(mw.tab_binary.pushButton_fast_prev, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 0

    # now label a little
    qtbot.mouseClick(mw.tab_binary.pushButton_fast_next, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 4
    qtbot.mouseClick(mw.tab_binary.pushButton_yes, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 5
    assert mw.tab_binary.label_score_prev.text() == "Yes"
    qtbot.mouseClick(mw.tab_binary.pushButton_no, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 6
    assert mw.tab_binary.label_score_prev.text() == "No"

    # go forward, label and then test fast_prev
    qtbot.mouseClick(mw.tab_binary.pushButton_next, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_next, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 8
    qtbot.mouseClick(mw.tab_binary.pushButton_no, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_no, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 10
    qtbot.mouseClick(mw.tab_binary.pushButton_fast_prev, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 7

    # go to the end and then test fast_next
    mw.tab_binary.goto_event(16)
    assert mw.tab_binary.event_index == 16
    qtbot.mouseClick(mw.tab_binary.pushButton_no, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_no, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_prev, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_binary.pushButton_prev, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 15
    qtbot.mouseClick(mw.tab_binary.pushButton_fast_next, QtCore.Qt.LeftButton)
    assert mw.tab_binary.event_index == 17


def test_session_load(qtbot, mw):
    """Load an .rtdc file with labeled data"""
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, True)
        dts.set_score("ml_score_r1u", 3, True)

        assert dts.get_scores_true(0) == ["ml_score_r1f"]
        assert dts.get_scores_true(1) == []
        assert dts.get_scores_true(2) == ["ml_score_r1f"]
        assert dts.get_scores_true(3) == ["ml_score_r1f", "ml_score_r1u"]
        assert dts.get_scores_true(4) == []

    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    idx = mw.tab_binary.comboBox_score.findData("ml_score_r1f")
    mw.tab_binary.comboBox_score.setCurrentIndex(idx)

    qtbot.mouseClick(mw.tab_binary.pushButton_start, QtCore.Qt.LeftButton)

    # The first event should be displayed, and it should be set to True
    assert not mw.tab_binary.pushButton_prev.isEnabled()
    assert mw.tab_binary.pushButton_yes.text() == "[Yes]"
    assert mw.tab_binary.pushButton_no.text() == "No"
    assert mw.tab_binary.label_score_next.text() == "nan"

    # Click the next button and check again
    qtbot.mouseClick(mw.tab_binary.pushButton_next, QtCore.Qt.LeftButton)
    assert mw.tab_binary.pushButton_prev.isEnabled()
    assert mw.tab_binary.pushButton_yes.text() == "Yes"
    assert mw.tab_binary.pushButton_no.text() == "No"
    assert mw.tab_binary.label_score_next.text() == "Yes"


@pytest.mark.parametrize("event_index,expected", [
    [-10, 0],
    [-1, 0],
    [0, 0],
    [16, 16],
    [17, 17],
    [18, 17],
    [5000, 17]])
def test_on_spin(event_index, expected, qtbot, mw):
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    # add spinbox
    qtbot.addWidget(mw.tab_binary.spinBox_jump_to)
    # set spinBox
    mw.tab_binary.spinBox_jump_to.setValue(event_index + 1)
    # check if event_index is updated correspondingly
    assert mw.tab_binary.event_index == expected


@pytest.mark.parametrize("event_index,expected", [
    [-10, 0],
    [-1, 0],
    [0, 0],
    [16, 16],
    [17, 17],
    [18, 17],
    [5000, 17]])
def test_update_spinBox(event_index, expected, qtbot, mw):
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # select binary tab
    mw.tabWidget.setCurrentIndex(1)
    # add spinbox
    qtbot.addWidget(mw.tab_binary.spinBox_jump_to)

    mw.tab_binary.goto_event(event_index)
    # check if spinBox is updated correspondingly
    assert mw.tab_binary.spinBox_jump_to.value() == expected + 1



import importlib.resources

from PyQt5 import QtWidgets, uic

import dclab


class TabSessionInfo(QtWidgets.QWidget):
    """Tab that displays .rtdc file DCTagSession information"""

    def __init__(self, *args, **kwargs):
        super(TabSessionInfo, self).__init__(*args, **kwargs)

        ref = importlib.resources.files("dctag.gui") / "tab_session.ui"
        with importlib.resources.as_file(ref) as path_ui:
            uic.loadUi(path_ui, self)

    def update_session(self, session):
        """Update this widget with the session info"""
        if not session:
            user = ""
            logs = "No session."
        else:
            user = session.user
            with session.score_lock:
                try:
                    with dclab.new_dataset(session.path) as ds:
                        logs = "\n".join(ds.logs["dctag-history"])
                except BaseException:
                    logs = f"Cannot get logs from '{session.path}'!"
        self.plainTextEdit_logs.setPlainText(logs)
        self.label_username.setText(user)
        self.label_num_sessions.setText(f"{logs.count('new session')}")



import pathlib
from unittest import mock

from PyQt5 import QtCore, QtWidgets
import pytest

from dctag import session
from .helper import get_clean_data_path


data_dir = pathlib.Path(__file__).parent / "data"


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)
    pass
    # A test function will be run at this point
    yield
    # Code that will run after your test
    # restore dctag-tester for other tests
    QtCore.QCoreApplication.setOrganizationName("MPL")
    QtCore.QCoreApplication.setOrganizationDomain("dc-cosmos.org")
    QtCore.QCoreApplication.setApplicationName("dctag")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setValue("user/name", "dctag-tester")
    QtWidgets.QApplication.processEvents(QtCore.QEventLoop.AllEvents, 3000)


def test_empty_session(mw):
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    # make sure things are disabled
    assert not mw.tab_multiple.isEnabled()


@pytest.mark.parametrize("event_index,expected", [
    [-10, 0],
    [-1, 0],
    [0, 0],
    [16, 16],
    [17, 17],
    [18, 17],
    [5000, 17]])
def test_goto_event_limits(event_index, expected, qtbot, mw):
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    mw.tab_multiple.comboBox_score.setItemChecked(0, True)  # r1f
    mw.tab_multiple.comboBox_score.setItemChecked(1, True)  # r1u
    assert mw.tab_multiple.comboBox_score.itemChecked(0)
    assert mw.tab_multiple.comboBox_score.itemChecked(1)
    assert not mw.tab_multiple.comboBox_score.itemChecked(2)
    # start labeling
    qtbot.mouseClick(mw.tab_multiple.pushButton_start, QtCore.Qt.LeftButton)

    # go to event
    mw.tab_multiple.goto_event(event_index)
    assert mw.tab_multiple.event_index == expected


def test_goto_event_button_labels(qtbot, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 1, False)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, False)
        dts.set_score("ml_score_r1u", 3, True)

    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    mw.tab_multiple.comboBox_score.setItemChecked(0, True)  # r1f
    mw.tab_multiple.comboBox_score.setItemChecked(1, True)  # r1u

    qtbot.mouseClick(mw.tab_multiple.pushButton_start, QtCore.Qt.LeftButton)

    assert len(mw.tab_multiple.label_buttons) == 2

    br1f, br1u = mw.tab_multiple.label_buttons
    assert br1f.pushButton.text() == "[R1F]"
    assert br1u.pushButton.text() == "!R1U"  # because it auto-filled!
    assert mw.tab_multiple.label_score_prev.text() == ""
    assert mw.tab_multiple.label_score_next.text() == "nan"

    # click on next.
    qtbot.mouseClick(mw.tab_multiple.pushButton_next, QtCore.Qt.LeftButton)

    assert br1f.pushButton.text() == "!R1F"
    assert br1u.pushButton.text() == "R1U"
    assert mw.tab_multiple.label_score_prev.text() == "R1F"
    assert mw.tab_multiple.label_score_next.text() == "R1F"

    # click on next.
    qtbot.mouseClick(mw.tab_multiple.pushButton_next, QtCore.Qt.LeftButton)

    assert br1f.pushButton.text() == "[R1F]"
    assert br1u.pushButton.text() == "!R1U"
    assert mw.tab_multiple.label_score_prev.text() == "nan"
    assert mw.tab_multiple.label_score_next.text() == "R1U"

    qtbot.mouseClick(mw.tab_multiple.pushButton_next, QtCore.Qt.LeftButton)

    assert br1f.pushButton.text() == "!R1F"
    assert br1u.pushButton.text() == "[R1U]"
    assert mw.tab_multiple.label_score_prev.text() == "R1F"
    assert mw.tab_multiple.label_score_next.text() == "nan"


def test_lock_in_twice(qtbot, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 1, False)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, False)
        dts.set_score("ml_score_r1u", 3, True)

    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    mw.tab_multiple.comboBox_score.setItemChecked(0, True)  # r1f
    mw.tab_multiple.comboBox_score.setItemChecked(1, True)  # r1u

    qtbot.mouseClick(mw.tab_multiple.pushButton_start, QtCore.Qt.LeftButton)

    assert len(mw.tab_multiple.label_buttons) == 2

    br1f, br1u = mw.tab_multiple.label_buttons
    assert br1f.pushButton.text() == "[R1F]"
    assert br1u.pushButton.text() == "!R1U"  # because it auto-filled!
    assert mw.tab_multiple.label_score_prev.text() == ""
    assert mw.tab_multiple.label_score_next.text() == "nan"

    # Now add another item and lock in again
    mw.tabWidget.setCurrentIndex(0)
    mw.tabWidget.setCurrentIndex(2)
    mw.tab_multiple.comboBox_score.setItemChecked(0, True)  # r1f
    mw.tab_multiple.comboBox_score.setItemChecked(1, True)  # r1u
    mw.tab_multiple.comboBox_score.setItemChecked(2, True)  # r20

    qtbot.mouseClick(mw.tab_multiple.pushButton_start, QtCore.Qt.LeftButton)

    br1f, br1u, br1n = mw.tab_multiple.label_buttons
    assert br1f.pushButton.text() == "[R1F]"
    assert br1u.pushButton.text() == "!R1U"  # because it auto-filled!
    assert br1n.pushButton.text() == "!R1N"  # because it auto-filled!


def test_event_push_buttons(qtbot, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester") as dts:
        dts.set_score("ml_score_r1f", 0, True)
        dts.set_score("ml_score_r1f", 1, False)
        dts.set_score("ml_score_r1u", 1, True)
        dts.set_score("ml_score_r1f", 2, True)
        dts.set_score("ml_score_r1f", 3, False)
        dts.set_score("ml_score_r1u", 3, True)

    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    mw.tab_multiple.comboBox_score.setItemChecked(0, True)  # r1f
    mw.tab_multiple.comboBox_score.setItemChecked(1, True)  # r1u

    qtbot.mouseClick(mw.tab_multiple.pushButton_start, QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 0
    qtbot.mouseClick(mw.tab_multiple.pushButton_next, QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 1
    qtbot.mouseClick(mw.tab_multiple.pushButton_fast_next,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 4
    qtbot.mouseClick(mw.tab_multiple.pushButton_prev, QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 3
    qtbot.mouseClick(mw.tab_multiple.pushButton_fast_prev,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 0

    # now label a little
    qtbot.mouseClick(mw.tab_multiple.pushButton_fast_next,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 4
    qtbot.mouseClick(mw.tab_multiple.label_buttons[0].pushButton,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 5
    assert mw.tab_multiple.label_score_prev.text() == "R1F"
    qtbot.mouseClick(mw.tab_multiple.label_buttons[1].pushButton,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 6
    assert mw.tab_multiple.label_score_prev.text() == "R1U"

    # go forward, label and then test fast_prev
    qtbot.mouseClick(mw.tab_multiple.pushButton_next, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_multiple.pushButton_next, QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 8
    qtbot.mouseClick(mw.tab_multiple.label_buttons[1].pushButton,
                     QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_multiple.label_buttons[1].pushButton,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 10
    qtbot.mouseClick(mw.tab_multiple.pushButton_fast_prev,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 7

    # go to the end and then test fast_next
    mw.tab_multiple.goto_event(16)
    assert mw.tab_multiple.event_index == 16
    qtbot.mouseClick(mw.tab_multiple.label_buttons[1].pushButton,
                     QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_multiple.label_buttons[1].pushButton,
                     QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_multiple.pushButton_prev, QtCore.Qt.LeftButton)
    qtbot.mouseClick(mw.tab_multiple.pushButton_prev, QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 15
    qtbot.mouseClick(mw.tab_multiple.pushButton_fast_next,
                     QtCore.Qt.LeftButton)
    assert mw.tab_multiple.event_index == 17


def test_start_without_events_checked(qtbot, monkeypatch, mw):
    path = get_clean_data_path()
    with session.DCTagSession(path, "dctag-tester"):
        pass

    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)

    with mock.patch.object(QtWidgets.QMessageBox, "warning") as mc:
        qtbot.mouseClick(mw.tab_multiple.pushButton_start,
                         QtCore.Qt.LeftButton)
        mc.assert_called_once()
    mw.close()


@pytest.mark.parametrize("event_index,expected", [
    [-10, 0],
    [-1, 0],
    [0, 0],
    [16, 16],
    [17, 17],
    [18, 17],
    [5000, 17]])
def test_on_spin(event_index, expected, qtbot, mw):
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    # add spinbox
    qtbot.addWidget(mw.tab_multiple.spinBox_jump_to)
    # set spinBox value
    mw.tab_multiple.spinBox_jump_to.setValue(event_index + 1)
    # check if event_index is updated correspondingly
    assert mw.tab_multiple.event_index == expected


@pytest.mark.parametrize("event_index,expected", [
    [-10, 0],
    [-1, 0],
    [0, 0],
    [16, 16],
    [17, 17],
    [18, 17],
    [5000, 17]])
def test_update_spinBox(event_index, expected, qtbot, mw):
    path = get_clean_data_path()
    # claim session
    with session.DCTagSession(path, "dctag-tester"):
        pass
    # open session
    mw.on_action_open(path)
    # select multiple tab
    mw.tabWidget.setCurrentIndex(2)
    # add spinbox
    qtbot.addWidget(mw.tab_multiple.spinBox_jump_to)

    mw.tab_multiple.goto_event(event_index)
    # check if spinBox is updated correspondingly
    assert mw.tab_multiple.spinBox_jump_to.value() == expected + 1



"""Handle DCTag sessions

A session in DCTag is nothing else than an .rtdc file. You
open an .rtdc file and DCTag automatically determines which
machine-learning features have previously been analyzed and
what could possibly happen next.
"""
import threading
import time
import pathlib
import warnings

import dclab
import h5py
import numpy as np

from ._version import version


class DCTagSessionClosedWarning(UserWarning):
    pass


class DCTagSessionError(BaseException):
    pass


class DCTagSessionClosedError(DCTagSessionError):
    pass


class DCTagSessionLockedError(DCTagSessionError):
    """Cannot load the session, because it is locked by another process"""


class DCTagSessionWriteError(DCTagSessionError):
    """Raised when it is not possible to write to a session"""


class DCTagSessionWrongUserError(DCTagSessionError):
    """Raised when the session user does not match"""
    def __init__(self, olduser, *args):
        self.olduser = olduser
        super(DCTagSessionWrongUserError, self).__init__(*args)


class DCTagSession:
    def __init__(self, path, user, linked_features=None, override_user=False):
        """Initialize a DCTag session

        Parameters
        ----------
        path: str or pathlib.Path
            Path to an .rtdc file used for labeling
        user: str
            Unique string (e.g. "Bambi") that identifies a user;
            The input file `path` will be bound to that username,
            making it impossible to edit the same .rtdc file using
            a different username.
        linked_features: list of str
            List for "ml_scores_" features that should be treated as
            linked when writing scores to disk. E.g. if you have the
            linked features 'ml_score_rbc' and 'ml_score_wbc', then
            the followin applies:

            - Setting the score of 'rbc' to True implies that the score
              of 'wbc' is False.
            - However, setting the score of 'rbc' to False, does not
              imply that the score of 'wbc' is True.

            Thus, if you are using this feature for labeling multiple
            scores, make sure to always only set True scores (so the
            other scores get set to False).
        override_user: bool
            Whether to override the `user` stored in the session.

        Notes
        -----
        Upon initialization this class creates a .dctag file with the
        same file name stem as `path` to indicate that a DCTag session
        is in progress. This can be thought of as a file lock. It is a
        precaution to prevent two people from working on the same file
        at the same time.

        The methods that alter the .rtdc file in this class are
        thread-safe (using `self.score_lock`).

        The design makes sure that the user can still write to the
        original .rtdc file, even if e.g. the original file is on a
        network share that has been remounted during rating.
        """
        #: Lock used internally to avoid writing to `history` and `scores`
        #: while saving data in `flush`
        self.score_lock = threading.Lock()
        #: Session path
        self.path = pathlib.Path(path)
        #: Lock-file for this session
        self.path_lock = self.path.with_suffix(".dctag")
        if self.path_lock.exists():
            raise DCTagSessionLockedError(
                f"Somebody else is currently working on {self.path} or "
                + "DCTag exited unexpectedly in a previous run! Please ask "
                + "Paul to implement session recovery!")
        #: Session user
        self.user = user.strip()
        # Whether session info has been written to the dctag-history log
        self._session_info_in_log_up_to_date = False
        # list of linked features (see self.linked_features)
        self._linked_features = []
        # claim this file
        self._claim_path(override_user=override_user)
        #: simple key-value dictionary of the current session history
        self.history = {}
        #: list of (feature, index, score) in the order set by the user
        self.scores = []
        #: scoring features that are linked for labeling
        self.linked_features = linked_features
        # determine length of the dataset
        with dclab.new_dataset(self.path) as ds:
            #: Number of events in the dataset
            self.event_count = len(ds)
        #: The internal scores cache is a dict with numpy arrays to keep
        #: track of all the scores for internal use only. This is not used
        #: for writing scores to .rtdc files. The scores cache is important
        #: for being able to keep working on a dataset when the underlying
        #: path is temporarily not available.
        self.scores_cache = {}
        with h5py.File(self.path, "a") as h5:
            # make a copy of all available scores in self.scores_cache
            for feat in h5["events"]:
                if feat.startswith("ml_score_") or feat.startswith("userdef"):
                    self.scores_cache[feat] = np.copy(h5["events"][feat])

        # finally, acquire the file system lock
        self.path_lock.touch()
        # keep track of whether we still have an open session
        self._closed = False

    def __bool__(self):
        """Convenience function; allows you to use `if session` case"""
        return not self._closed

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def _claim_path(self, override_user=False):
        """Attribute this file to self.user"""
        with h5py.File(self.path, "a") as h5:
            hw = dclab.RTDCWriter(h5, mode="append")
            h5.require_group("logs")
            dctag_history = "dctag-history"
            if len(h5["logs"].get(dctag_history, ["placeholder"])) == 0:
                # remove empty logs
                del h5["logs"][dctag_history]

            if dctag_history not in h5["logs"]:
                hw.store_log(dctag_history, f"user: {self.user}")
            else:
                # Check whether the user in the file matches us
                h5userstr = h5["logs"][dctag_history][0]
                if isinstance(h5userstr, bytes):
                    h5userstr = h5userstr.decode("utf-8")
                if h5userstr.startswith("user:"):
                    h5user = h5userstr.split(":")[1].strip()
                    if h5user != self.user:
                        if override_user:
                            h5["logs"][dctag_history][0] = f"user: {self.user}"
                            hw.store_log(dctag_history,
                                         f"Session force-claimed from "
                                         f"{h5user} by {self.user}.")
                        else:
                            raise DCTagSessionWrongUserError(
                                h5user,
                                f"Expected user '{self.user}' in "
                                + f"'{self.path}', got '{h5user}'!")
                else:
                    # Something went wrong (maybe lost history).
                    # Reinstate the claim!
                    h5["logs"]["dctag-history"][0] = f"user: {self.user}"

    @property
    def linked_features(self):
        return self._linked_features

    @linked_features.setter
    def linked_features(self, linked_features):
        # Acquire a score_lock, because labeling might go on
        # in another thread, and we want the correct history.
        with self.score_lock:
            linked_features = linked_features or []
            if self._linked_features == linked_features:
                # nothing to do
                pass
            else:
                # write the scores and history now
                self.write_history(clear_history=True)
                self.write_scores(clear_scores=True)
                # make sure the session info is written to the logs
                # in the next call to self.write_history
                self._session_info_in_log_up_to_date = False
                # finally, set the linked features internally
                self._linked_features = sorted(linked_features)

    def assert_session_open(self, purpose="perform an undefined task",
                            strict=False):
        """Warn or raise an error if the session is closed

        Parameters
        ----------
        purpose: str
            Describe why you need the session open; used for debugging.
        strict: bool
            Whether to force raising a DCTagSessionClosedError; otherwise
            a DCTagSessionClosedWarning may be raised in situations where
            `self.history` or `self.scores` are empty.
        """
        if self._closed:
            if self.history or self.scores or strict:
                raise DCTagSessionClosedError(
                    "The session has been closed, but there are still data to "
                    + f"be written to '{self.path}'! Cannot {purpose}.",)
            else:
                warnings.warn(
                    "Session has been closed, but you are trying to "
                    + f"{purpose} which requires an open session. Luckily, "
                    + "there is nothing that needs to be written to disk, but "
                    + "you should try to avoid this anyway.",
                    DCTagSessionClosedWarning)

    def autocomplete_linked_features(self):
        """Autocomplete False for linked features"""
        with self.score_lock:
            # Create a concatenated array with all current scores
            fscores = np.zeros((self.event_count, len(self.linked_features)),
                               dtype=float)
            for ii, feat in enumerate(self.linked_features):
                fscores[:, ii] = self.require_dict_score_dataset(
                    self.scores_cache, feat)
            # Sanity check
            if np.any(np.nansum(fscores, axis=1) > 1):
                raise ValueError(
                    f"Some of the scores {self.linked_features} in "
                    + f"{self.path} have ambiguous labels! Make sure that "
                    + "always only one of those scores is labeled as True/Yes."
                    )
            # We are safe
            for ii, feat in enumerate(self.linked_features):
                mask_true = self.scores_cache[feat] == 1
                for other_feat in self.linked_features:
                    if other_feat != feat:
                        mask_nan = np.isnan(self.scores_cache[other_feat])
                        idx_new = np.where(
                            np.logical_and(mask_true, mask_nan))[0]
                        for idx in idx_new:
                            self.scores.append((other_feat, idx, False))
                            self.scores_cache[other_feat][idx] = False

    def backup_scores(self, path):
        """Backup current scores in an HDF5 file

        This can be used as a last resort to save score data if
        the original `self.path` has gone away for some reason.
        """
        with h5py.File(path, mode="w") as h5:
            with self.score_lock:
                for feat in self.scores_cache:
                    h5[feat] = self.scores_cache[feat]
                h5.attrs["path_original"] = str(self.path)

    def close(self):
        """Close this session, flushing everything to `self.path`"""
        self.flush()
        with self.score_lock:
            # call this function in the score_lock context again to
            # be on the safe side.
            self.assert_session_open("close the session")
            self._closed = True
            self.path_lock.unlink(missing_ok=True)

    def flush(self):
        """Flush all changes made to disk

        You should call this method regularly. It is thread-safe,
        so you may call it in regular intervals using a background
        thread.
        """
        with self.score_lock:
            self.assert_session_open("flush the session")
            try:
                self.write_scores(clear_scores=True)
                self.write_history(clear_history=True)
            except BaseException as exc:
                raise DCTagSessionWriteError(
                    f"Could not write to session {self.path}!") from exc

    def get_score(self, feature, index):
        """Return the score of a specific feature at that index

        Parameters
        ----------
        feature: str
            Name of the machine-learning feature (e.g. "ml_score_buk")
        index: int
            Event index (starts at 0)

        Returns
        -------
        score: bool or np.nan
            The score value (nan if not defined)

        Notes
        -----
        This method is thread-safe.
        """
        # We use the score cache for that
        with self.score_lock:
            self.assert_session_open(f"get the score {feature} at {index}")
            if feature not in self.scores_cache:
                value = np.nan
            else:
                value = self.scores_cache[feature][index]
                if not np.isnan(value):
                    value = bool(round(value))
            return value

    def get_scores_true(self, index):
        """Return the feature names that are labeld True for one event

        Parameters
        ----------
        index: int
            Event index (starts at 0)

        Returns
        -------
        features: list of str
            Feature names
        """
        true_features = []
        for feature in self.scores_cache:
            if self.get_score(feature, index) is True:
                true_features.append(feature)
        return sorted(true_features)

    def reset_score(self, feature, index, reset_linked=True):
        """Set the score at `index` to `np.nan`

        Parameters
        ----------
        feature: str
            Name of the machine-learning feature (e.g. "ml_score_buk")
        index: int
            Event index (starts at 0)
        reset_linked: bool
            Also reset all linked features if `feature` in
            `self.linked_features`.
        """
        if reset_linked and feature in self.linked_features:
            # Recurse one level and reset all features
            for feat in self.linked_features:
                self.reset_score(feat, index, reset_linked=False)
        else:
            # Do the actual resetting
            with self.score_lock:
                self.assert_session_open(
                    f"reset the score {feature} at {index}", strict=True)
                # scores list
                self.scores.append((feature, index, np.nan))

                # history list
                # (Note that this count value may be larger than the actual
                # updated number of events of the ml_score, because `feat_list`
                # may have multiple entries with the same index. This is OK).
                key = f"{feature} count reset"
                self.history.setdefault(key, 0)
                self.history[key] += 1

                self.require_dict_score_dataset(self.scores_cache, feature)
                self.scores_cache[feature][index] = np.nan

    def set_score(self, feature, index, value):
        """Set the feature score of an event in the current dataset

        Parameters
        ----------
        feature: str
            Name of the machine-learning feature (e.g. "ml_score_buk")
        index: int
            Event index (starts at 0)
        value: bool
            Boolean value indicating whether the event
            belongs to the `feature` class

        Notes
        -----
        This method is thread-safe.
        """
        if (not (feature.startswith("userdef")
                 or (feature.startswith("ml_score_") and
                     len(feature) == len("ml_score_???")))):
            raise ValueError(
                "Expected 'ml_score_xxx' or 'userdef*' feature, "
                + f"got '{feature}'!")
        with self.score_lock:
            self.assert_session_open(f"set the score {feature} at {index}",
                                     strict=True)
            # scores list
            self.scores.append((feature, index, value))

            # history list
            # (Note that this count value may be larger than the actual
            # updated number of events of the ml_score, because `feat_list`
            # may have multiple entries with the same index. This is OK).
            key = f"{feature} count {value}"
            self.history.setdefault(key, 0)
            self.history[key] += 1

            for feat in self.linked_features:
                self.require_dict_score_dataset(self.scores_cache, feat)
            self.require_dict_score_dataset(self.scores_cache, feature)

            self.scores_cache[feature][index] = value
            self.populate_linked_features(
                feature=feature,
                index=index,
                value=value,
                linked_feature_dict=self.scores_cache)

    def write_history(self, clear_history=False):
        """Write accomplishments to the history log in `self.path`

        The history log is a human-readable summary of the changes
        made in a session.

        Parameters
        ----------
        clear_history: bool
            Whether to clear `self.history`. Only set to True if you
            have previously acquired `self.score_lock`!

        Notes
        -----
        This method is NOT thread-safe. Use `self.flush` instead!
        """
        if self.history:
            date = time.strftime("%Y-%m-%d %H:%M:%S")
            with dclab.RTDCWriter(self.path, mode="append") as hw:
                if not self._session_info_in_log_up_to_date:
                    hw.store_log(
                        "dctag-history",
                        ["",
                         f"{date} New session with DCTag {version}",
                         f"{date} Linked features: {self.linked_features}"
                         ])
                for key in sorted(self.history.keys()):
                    hw.store_log("dctag-history",
                                 f"{date} {key}: {self.history[key]}")
            if clear_history:
                # clear history
                self.history.clear()

    def write_scores(self, clear_scores=False):
        """Write the machine-learning scores to `self.path`

        Parameters
        ----------
        clear_scores: bool
            Whether to clear `self.scores`. Only set to True if you
            have previously acquired `self.score_lock`!

        Notes
        -----
        This method is NOT thread-safe. Use `self.flush` instead!
        """
        if self.scores:
            with h5py.File(self.path, mode="r+") as h5:
                # make sure that all linked features are available
                for feat in self.linked_features:
                    self.require_h5_score_dataset(h5, feat)
                # populate features
                for feat, idx, val in self.scores:
                    sc_ds = self.require_h5_score_dataset(h5, feat)
                    sc_ds[idx] = val
                    # Write False to the other linked features
                    self.populate_linked_features(
                        feature=feat,
                        index=idx,
                        value=val,
                        linked_feature_dict=h5["events"])

            if clear_scores:
                self.scores.clear()

    def require_h5_score_dataset(self, h5, feature):
        """Return dataset in the `h5["events"]` group for `feature`"""
        if feature not in h5["events"]:
            # create a nan-filled dataset for this feature
            data = np.zeros(self.event_count, dtype=float) * np.nan
            h5["events"].create_dataset(feature, data=data)
        return h5["events"][feature]

    def require_dict_score_dataset(self, ndict, feature):
        """Return dataset in `ndict` for `feature`"""
        # internal score cache
        if feature not in ndict:
            ndict[feature] = np.zeros(self.event_count, dtype=float) * np.nan
        return ndict[feature]

    def populate_linked_features(self, feature, index, value,
                                 linked_feature_dict):
        """Set values in linked_feature_dict to False if value is True

        Notes
        -----
        We assume that all `self.linked_features` are in
        `linked_feature_dict`. This is done during `__init__`.
        """
        if value is True and feature in self.linked_features:
            for feat in self.linked_features:
                if feat != feature:
                    ln_sc_ds = linked_feature_dict[feat]
                    ln_sc_ds[index] = False


def is_dctag_session(path):
    """Return True if `path` has a dctag-history log"""
    with h5py.File(path, "r") as h5:
        return "logs/dctag-history" in h5
